////////////////////////////////////////////////////////////////
// RSSP - Resize Sprit data
//                         Developed by NEY / KYOSKE 1998-1999

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "rsheap.h"
#include "rscnf.h"

#ifdef WIN32
#include <io.h>
#else
#define __DOS_INLINE__
#include <sys/dos.h>
#endif

#define  SRCSIZE  (514*514)
#define  DSTSIZE  (514*514)
#define  PALSIZE  16
#define  GRPSIZE  16
#define  RESO1     8
#define  RESO2     6
#define  COLORBIT  8

#define  MD_SP_FILE   0
#define  MD_SP_DIR    1
#define  MD_BMP_FILE  2
#define  MD_BMP_DIR   3
#define  MD_VRAM      4
#define  MD_VRAM1     5
#define  MD_VRAM2     6
#define  MD_VRAM3     7
#define  MD_VRAM4     8

#define  MD_PAL_FILE  0

#define  COPYRIGHT  "RSSP ver1.10  XvCgf[^kc[  Developed by NEY / KYOSKE 1998-1999\n"

char *pConfig = "rssp.cnf";

short md_input;
short md_output;
short md_palette;
short md_float;

unsigned char *src;
unsigned char *dst;

short srcw;  // src 
short srch;  // src 
short sbpl;  // src byte per line

short dstw;  // dst 
short dsth;  // dst 
short dbpl;  // dst byte per line

double hscale;  // ̏k
double vscale;  // c̏k

short weight1[PALSIZE][GRPSIZE];
short weight2[PALSIZE][GRPSIZE];
short palette[GRPSIZE][PALSIZE];
short npal[GRPSIZE];
short red[PALSIZE];
short green[PALSIZE];
short blue[PALSIZE];
char *gname[GRPSIZE];
short ngrp;

int *multbl;

#define colorDistance(r1, g1, b1, r2, g2, b2) \
	((*(multbl+(r1)-(r2)))+(*(multbl+(g1)-(g2)))+(*(multbl+(b1)-(b2))))
//	(*(multbl+((((r1)*5+(g1)*9+(b1)*2)-((r2)*5+(g2)*9+(b2)*2))>>4)))
#define fraction(n) ((n)-(int)((n)+0.000000000000005))

void Convert(char *infile, char *outfile);
void Resize1();
void Resize2();
void Resize3();
void SetScale();
void SetPriority();
void SetWeight();
void InitMultbl();
void ReadPalette(char *name);
void ReadGPalette();
void ReadSprite(char *name, unsigned char *buf, short *pdx, short *pdy, short *pbpl);
void WriteSprite(char *name, unsigned char *buf, short dx, short dy, short bpl);
void ReadBitmap(char *name, unsigned char *buf, short *pdx, short *pdy, short *pbpl);
void WriteBitmap(char *name, unsigned char *buf, short dx, short dy, short bpl);
void ReadVram(short mode, unsigned char *buf, short *pdx, short *pdy, short *pbpl);
void WriteVram(short mode, unsigned char *buf, short dx, short dy, short bpl);
char *FindFirst(char *dir, char *ext);
char *FindNext();
void FindClose();
char *NextFileName(char *infile, char *outfile);

void main(int argc, char *argv[]) {
	char *infile;
	char *outfile;
	char *palfile;
	char *in;
	char *out;

	printf(COPYRIGHT);

	if (argc > 1) pConfig = argv[1];

	HeapInit(1024*1024);  // 1M byte
	CnfInit();
	if (CnfLoad(pConfig) == FALSE) {
		printf("Error: can't open %s!\n", pConfig);
		exit(1);
	}

	src = (unsigned char *)HeapAlloc(SRCSIZE);
	dst = (unsigned char *)HeapAlloc(DSTSIZE);
	InitMultbl();

	md_input   = CnfGetInt(pConfig, "mode", "input");
	md_output  = CnfGetInt(pConfig, "mode", "output");
	md_palette = CnfGetInt(pConfig, "mode", "palette");
	md_float   = CnfGetInt(pConfig, "mode", "float");

	infile  = CnfGetStr(pConfig, "file", "input");
	outfile = CnfGetStr(pConfig, "file", "output");
	palfile = CnfGetStr(pConfig, "file", "palette");

	SetScale();
	SetPriority();
	SetWeight();

	if (md_palette == MD_PAL_FILE) {
		ReadPalette(palfile);
	} else if (md_palette == MD_VRAM) {
		ReadGPalette();
	}

	if (md_input == MD_SP_DIR) {
		in = FindFirst(infile, "*.sp");
		while (in) {
			printf("%s\n", in);
			out = NextFileName(in, outfile);
			Convert(in, out);
			in = FindNext();
		}
		FindClose();
	} else if (md_input == MD_BMP_DIR) {
		in = FindFirst(infile, "*.bmp");
		while (in) {
			printf("%s\n", in);
			out = NextFileName(in, outfile);
			Convert(in, out);
			in = FindNext();
		}
		FindClose();
	} else {
		printf("%s\n", infile);
		out = NextFileName(infile, outfile);
		Convert(infile, out);
	}

	CnfTerm();
	HeapTerm();
}

void Convert(char *infile, char *outfile) {
	switch (md_input) {
	case MD_SP_FILE:
	case MD_SP_DIR:
		ReadSprite(infile, src, &srcw, &srch, &sbpl);
		break;
	case MD_BMP_FILE:
	case MD_BMP_DIR:
		ReadBitmap(infile, src, &srcw, &srch, &sbpl);
		break;
	case MD_VRAM:
		ReadVram(0, src, &srcw, &srch, &sbpl);
		break;
	case MD_VRAM1:
		ReadVram(1, src, &srcw, &srch, &sbpl);
		break;
	case MD_VRAM2:
		ReadVram(2, src, &srcw, &srch, &sbpl);
		break;
	case MD_VRAM3:
		ReadVram(3, src, &srcw, &srch, &sbpl);
		break;
	case MD_VRAM4:
		ReadVram(4, src, &srcw, &srch, &sbpl);
		break;
	default:
		printf("Error: no suport input mode!\n");
		exit(1);
	}

	switch (md_float) {
	case 0:  // Œ菬_[h
		Resize1();
		break;
	case 1:  // Œ{_[h
		Resize2();
		break;
	case 2:  // _[h
		Resize3();
		break;
	}

	switch (md_output) {
	case MD_SP_FILE:
	case MD_SP_DIR:
		WriteSprite(outfile, dst, srcw, srch, dbpl);
		break;
	case MD_BMP_FILE:
	case MD_BMP_DIR:
		WriteBitmap(outfile, dst, srcw, srch, dbpl);
		break;
	case MD_VRAM:
		WriteVram(0, dst, srcw, srch, dbpl);
		break;
	case MD_VRAM1:
		WriteVram(1, dst, srcw, srch, dbpl);
		break;
	case MD_VRAM2:
		WriteVram(2, dst, srcw, srch, dbpl);
		break;
	case MD_VRAM3:
		WriteVram(3, dst, srcw, srch, dbpl);
		break;
	case MD_VRAM4:
		WriteVram(4, dst, srcw, srch, dbpl);
		break;
	default:
		printf("Error: no suport output mode!\n");
		exit(1);
	}
}

void Resize1() {
	int wt[GRPSIZE];
	int hr[GRPSIZE];
	int hg[GRPSIZE];
	int hb[GRPSIZE];
	int px[GRPSIZE];
	unsigned char *p1, *p2;
	short pal,  pal2;
	int dist, dist2;
	short st;
	short dx, dy;
	int   ex, ey, nx, ny, x, y, n;
	int   r, g, b;
	short w, h, i;
#ifndef WIN32
	char  s[16];
	short line;
#endif

	dstw = (int)(srcw * hscale + 0.5);
	dsth = (int)(srch * vscale + 0.5);
	dbpl = sbpl;

	dx = (short)(hscale * (1<<RESO1) + 0.5);
	dy = (short)(vscale * (1<<RESO1) + 0.5);
	ex = dx * (1<<RESO1);
	ey = dy * (1<<RESO1);

#ifndef WIN32
	line = 0;
#endif
	p1 = dst;
	y  = 0;
	while (y < ey) {
		p2 = p1;
		x  = 0;
		while (x < ex) {
			// d݂ƐF̏Wv
			for (i = 0; i < ngrp; i++) {
				wt[i] = 0;
				hr[i] = 0;
				hg[i] = 0;
				hb[i] = 0;
				px[i] = 0;
			}
			h = 0;
			while (h < (1<<RESO1)) {
				ny = h + dy - ((y+h) % dy);
				if (ny > (1<<RESO1)) ny = (1<<RESO1);
				w = 0;
				while (w < (1<<RESO1)) {
					nx = w + dx - ((x+w) % dx);
					if (nx > (1<<RESO1)) nx = (1<<RESO1);
					n = ((nx - w) * (ny - h)) >> ((RESO1-RESO2)+(RESO1-RESO2));
					pal = src[((y+h)/dy)*sbpl+((x+w)/dx)];
					for (i = 0; i < ngrp; i++) {
						wt[i] += weight1[pal][i] * n;
						if (weight1[pal][i] > 0) {
							hr[i] += red[pal]   * n;
							hg[i] += green[pal] * n;
							hb[i] += blue[pal]  * n;
							px[i] += n;
						}
					}
					w = nx;
				}
				h = ny;
			}

			// \O[v𒲂ׂ
			st = 0;
			for (i = 1; i < ngrp; i++) {
				if (wt[i] > wt[st]) st = i;
			}

			// \O[v\FI
			if (npal[st] > 1) {
				r = hr[st] / px[st];
				g = hg[st] / px[st];
				b = hb[st] / px[st];
				pal  = palette[st][0];
				dist = colorDistance(r, g, b, red[pal], green[pal], blue[pal]);
				dist /= weight2[pal][st];
				for (i = 1; i < npal[st]; i++) {
					pal2  = palette[st][i];
					dist2 = colorDistance(r, g, b, red[pal2], green[pal2], blue[pal2]);
					dist2 /= weight2[pal2][st];
					if (dist2 < dist) {
						pal  = pal2;
						dist = dist2;
					}
				}
			} else {
				pal = palette[st][0];
			}

			// \Fݒ
			*(p2++) = (unsigned char)pal;
			x += (1<<RESO1);
		}
		p1 += dbpl;
		y  += (1<<RESO1);
#ifndef WIN32
		line++;
		sprintf(s, "line %4d/%4d", line, dsth);
		_dos_print("\033[1K\015");
		_dos_print(s);
#endif
	}
#ifndef WIN32
	_dos_print("\033[1K\015");
#endif
}

void Resize2() {
	int wt[GRPSIZE];
	int hr[GRPSIZE];
	int hg[GRPSIZE];
	int hb[GRPSIZE];
	int px[GRPSIZE];
	int xtbl[514];
	int ytbl[514];
	unsigned char *p1, *p2;
	short pal,  pal2;
	int dist, dist2;
	short st;
	int   sx, sy, ex, ey, nx, ny, x, y, w, h, n;
	int   r, g, b;
	short i;
#ifndef WIN32
	char  s[16];
	short line;
#endif

	dstw = (int)(srcw * hscale + 0.5);
	dsth = (int)(srch * vscale + 0.5);
	dbpl = sbpl;

	for (x = 0; x <= dstw; x++) {
		xtbl[x] = (int)((x<<RESO1) / hscale + 0.5);
	}
	for (y = 0; y <= dsth; y++) {
		ytbl[y] = (int)((y<<RESO1) / vscale + 0.5);
	}

#ifndef WIN32
	line = 0;
#endif
	p1 = dst;
	y  = 0;
	while (y < dsth) {
		sy = ytbl[y  ];
		ey = ytbl[y+1];

		p2 = p1;
		x  = 0;
		while (x < dstw) {
			sx = xtbl[x  ];
			ex = xtbl[x+1];

			// d݂ƐF̏Wv
			for (i = 0; i < ngrp; i++) {
				wt[i] = 0;
				hr[i] = 0;
				hg[i] = 0;
				hb[i] = 0;
				px[i] = 0;
			}
			h = sy;
			while (h < ey) {
				ny = h + (1<<RESO1) - (h & ((1<<RESO1)-1));
				if (ny > ey) ny = ey;
				w = sx;
				while (w < ex) {
					nx = w + (1<<RESO1) - (w & ((1<<RESO1)-1));
					if (nx > ex) nx = ex;
					n = ((nx - w) * (ny - h)) >> ((RESO1-RESO2)+(RESO1-RESO2));
					pal = src[(h>>RESO1)*sbpl+(w>>RESO1)];
					for (i = 0; i < ngrp; i++) {
						wt[i] += weight1[pal][i] * n;
						if (weight1[pal][i] > 0) {
							hr[i] += red[pal]   * n;
							hg[i] += green[pal] * n;
							hb[i] += blue[pal]  * n;
							px[i] += n;
						}
					}
					w = nx;
				}
				h = ny;
			}

			// \O[v𒲂ׂ
			st = 0;
			for (i = 1; i < ngrp; i++) {
				if (wt[i] > wt[st]) st = i;
			}

			// \O[v\FI
			if (npal[st] > 1) {
				r = hr[st] / px[st];
				g = hg[st] / px[st];
				b = hb[st] / px[st];
				pal  = palette[st][0];
				dist = colorDistance(r, g, b, red[pal], green[pal], blue[pal]);
				dist /= weight2[pal][st];
				for (i = 1; i < npal[st]; i++) {
					pal2  = palette[st][i];
					dist2 = colorDistance(r, g, b, red[pal2], green[pal2], blue[pal2]);
					dist2 /= weight2[pal2][st];
					if (dist2 < dist) {
						pal  = pal2;
						dist = dist2;
					}
				}
			} else {
				pal = palette[st][0];
			}

			// \Fݒ
			*(p2++) = (unsigned char)pal;
			x++;
		}
		p1 += dbpl;
		y++;
#ifndef WIN32
		line++;
		sprintf(s, "line %4d/%4d", line, dsth);
		_dos_print("\033[1K\015");
		_dos_print(s);
#endif
	}
#ifndef WIN32
	_dos_print("\033[1K\015");
#endif
}

void Resize3() {
	double wt[GRPSIZE];
	double hr[GRPSIZE];
	double hg[GRPSIZE];
	double hb[GRPSIZE];
	double px[GRPSIZE];
	double xtbl[514];
	double ytbl[514];
	unsigned char *p1, *p2;
	short pal,  pal2;
	int dist, dist2;
	short st;
	int   x, y;
	double ex, ey, nx, ny, sx, sy, w, h, n;
	int    r, g, b;
	short  i;
#ifndef WIN32
	char  s[16];
	short line;
#endif

	dstw = (int)(srcw * hscale + 0.5);
	dsth = (int)(srch * vscale + 0.5);
	dbpl = sbpl;

	for (x = 0; x <= dstw; x++) {
		xtbl[x] = x / hscale;
	}
	for (y = 0; y <= dsth; y++) {
		ytbl[y] = y / vscale;
	}

#ifndef WIN32
	line = 0;
#endif
	p1 = dst;
	y  = 0;
	while (y < dsth) {
		sy = ytbl[y  ];
		ey = ytbl[y+1];

		p2 = p1;
		x  = 0;
		while (x < dstw) {
			sx = xtbl[x  ];
			ex = xtbl[x+1];

			// d݂ƐF̏Wv
			for (i = 0; i < ngrp; i++) {
				wt[i] = 0.0;
				hr[i] = 0.0;
				hg[i] = 0.0;
				hb[i] = 0.0;
				px[i] = 0.0;
			}
			h  = sy;
			while (h < ey) {
				ny = h + 1.0 - fraction(h);
				if (ny > ey) ny = ey;
				w = sx;
				while (w < ex) {
					nx = w + 1.0 - fraction(w);
					if (nx > ex) nx = ex;
					n = (nx - w) * (ny - h);
					pal = src[((int)h)*sbpl+((int)w)];
					for (i = 0; i < ngrp; i++) {
						wt[i] += weight1[pal][i] * n;
						if (weight1[pal][i] > 0) {
							hr[i] += red[pal]   * n;
							hg[i] += green[pal] * n;
							hb[i] += blue[pal]  * n;
							px[i] += n;
						}
					}
					w = nx;
				}
				h = ny;
			}

			// \O[v𒲂ׂ
			st = 0;
			for (i = 1; i < ngrp; i++) {
				if (wt[i] > wt[st]) st = i;
			}

			// \O[v\FI
			if (npal[st] > 1) {
				r = (int)(hr[st] / px[st]);
				g = (int)(hg[st] / px[st]);
				b = (int)(hb[st] / px[st]);
				pal  = palette[st][0];
				dist = colorDistance(r, g, b, red[pal], green[pal], blue[pal]);
				dist /= weight2[pal][st];
				for (i = 1; i < npal[st]; i++) {
					pal2  = palette[st][i];
					dist2 = colorDistance(r, g, b, red[pal2], green[pal2], blue[pal2]);
					dist2 /= weight2[pal2][st];
					if (dist2 < dist) {
						pal  = pal2;
						dist = dist2;
					}
				}
			} else {
				pal = palette[st][0];
			}

			// \Fݒ
			*(p2++) = (unsigned char)pal;
			x++;
		}
		p1 += dbpl;
		y++;
#ifndef WIN32
		line++;
		sprintf(s, "line %4d/%4d", line, dsth);
		_dos_print("\033[1K\015");
		_dos_print(s);
#endif
	}
#ifndef WIN32
	_dos_print("\033[1K\015");
#endif
}

void SetScale() {
	char *p;
	int m, d;

	// 
	p = CnfGetStr(pConfig, "scale", "horz");
	if (p) {
		m = 0;
		d = 1;
		while (*p && isdigit(*p)) m = m*10 + *(p++) - '0';
		if (*p == '.') {
			p++;
			while (*p && isdigit(*p)) {
				m  = m*10 + *(p++) - '0';
				d *= 10;
			}
		}
		if (d != 1) {
			hscale = ((double)m) / ((double)d);
		} else {
			hscale = (double)m;
		}
	} else {
		hscale = 1.0;
	}

	// 
	p = CnfGetStr(pConfig, "scale", "vert");
	if (p) {
		m = 0;
		d = 1;
		while (*p && isdigit(*p)) m = m*10 + *(p++) - '0';
		if (*p == '.') {
			p++;
			while (*p && isdigit(*p)) {
				m  = m*10 + *(p++) - '0';
				d *= 10;
			}
		}
		if (d != 1) {
			vscale = ((double)m) / ((double)d);
		} else {
			vscale = (double)m;
		}
	} else {
		vscale = 1.0;
	}
}

void SetPriority() {
	ITEM *item;
	char *p;
	short i, x;

	i = 0;
	item = CnfFindItem(pConfig, "priority");
	while (item) {
		p = item->str;
		if (p) {
			x = 0;
			while (*p) {
				if (isxdigit(*p)) {
					if (isupper(*p)) {
						palette[i][x] = *p - 'A' + 10;
					} else if (islower(*p)) {
						palette[i][x] = *p - 'a' + 10;
					} else {
						palette[i][x] = *p - '0';
					}
					x++;
				}
				p++;
			}
			gname[i] = item->name;
			npal[i]  = x;
			i++;
		}
		item = item->next;
	}
	ngrp = i;
}

void SetWeight() {
	short i, x, n, w;

	// O[vԂ̏d
	for (i = 0; i < ngrp; i++) {
		char *p = CnfGetStr(pConfig, "weight1", gname[i]);
		n = npal[i];

		for (x = 0; x < 16; x++) weight1[x][i] = 0;

		w = 1;
		x = 0;
		if (p) {
			while (*p && (x < n)) {
				while (isspace(*p)) p++;
				if (isdigit(*p) == FALSE) break;

				w = 0;
				while (isdigit(*p)) {
					w = w * 10 + *(p++) - '0';
				}
				weight1[palette[i][x++]][i] = w;

				while (isspace(*p)) p++;

				if (*p == ',') p++;
			}
		}
		while (x < n) {
			weight1[palette[i][x++]][i] = w;
		}
	}

	// O[vpbgFԂ̏d
	for (i = 0; i < ngrp; i++) {
		char *p = CnfGetStr(pConfig, "weight2", gname[i]);
		n = npal[i];

		for (x = 0; x < 16; x++) weight2[x][i] = 0;

		w = 1;
		x = 0;
		if (p) {
			while (*p && (x < n)) {
				while (isspace(*p)) p++;
				if (isdigit(*p) == FALSE) break;

				w = 0;
				while (isdigit(*p)) {
					w = w * 10 + *(p++) - '0';
				}
				weight2[palette[i][x++]][i] = w;

				while (isspace(*p)) p++;

				if (*p == ',') p++;
			}
		}
		while (x < n) {
			weight2[palette[i][x++]][i] = w;
		}
	}
}

void InitMultbl() {
	short i, n;

	multbl = ((int*)HeapAlloc(sizeof(int)*512)) + 256;
	n = (1<<COLORBIT)-1;

	for (i = -n; i <= n; i++) {
		*(multbl+i) = i*i;
	}
}

void ReadPalette(char *name) {
	FILE *fp;
	char buf[32];
	int   d;
	short i;

	// t@CJ
	fp = fopen(name, "rb");
	if (fp == NULL) {
		printf("Error: can't open %s!\n", name);
		return;
	}

	// f[^ǂݍ
	fread(buf, sizeof(char), 32, fp);

	// pbgݒ肷
	for (i = 0; i < 16; i++) {
		d = ((((int)buf[i*2] & 0xff)) << 8) | (((int)buf[i*2+1]) & 0xff);
		red[i]   = ((d >> 3) & 0xf8) | ((d >>  8) & 0x07);
		green[i] = ((d >> 8) & 0xf8) | ((d >> 13) & 0x07);
		blue[i]  = ((d << 2) & 0xf8) | ((d >>  3) & 0x07);
	}

	// t@C
	fclose(fp);
}

void ReadGPalette() {
#ifdef WIN32
	printf("Error: G-Palette read not saport for WIN32!\n");
	exit(1);
#else
	short *p = (short*)(void*)0xE82000;
	short d, i;

	for (i = 0; i < 16; i++) {
		d = *(p++);
		red[i]   = ((d >> 3) & 0xf8) | ((d >>  8) & 0x07);
		green[i] = ((d >> 8) & 0xf8) | ((d >> 13) & 0x07);
		blue[i]  = ((d << 2) & 0xf8) | ((d >>  3) & 0x07);
	}
#endif
}

void ReadSprite(char *name, unsigned char *buf, short *pdx, short *pdy, short *pbpl) {
	FILE *fp;
	static int tblx[] = {0, 0, 8, 8};
	static int tbly[] = {0, 8, 0, 8};
	unsigned char tmp[128];
	unsigned char *p;
	unsigned long d;
	short dx, dy, bpl, bpi;
	int x, y, tx, ty, w, h, i;

	// t@CJ
	fp = fopen(name, "rb");
	if (fp == NULL) {
		printf("Error: can't open %s!\n", name);
		return;
	}

	// p[^̐ݒ(̂P)
	dx  = 256;
	dy  = 256;
	bpl = dx + 2;
	bpi = (dy + 2) * bpl;

	// f[^ϊēǂݍ
	for (y = 0; y < dy; y += 16) {
		for (x = 0; x < dx; x += 16) {
			fread(tmp, sizeof(char), 128, fp);
			p = tmp;
			for (i = 0; i < 4; i++) {
				tx = tblx[i];
				ty = tbly[i];
				for (h = 0; h < 8; h++) {
					d  = ((int)(*(p++))) << 24;
					d |= ((int)(*(p++))) << 16;
					d |= ((int)(*(p++))) <<  8;
					d |= ((int)(*(p++)));
					for (w = 7; w >= 0; w--) {
						buf[(y+ty+h)*bpl+(x+tx+w)] = (unsigned char)(d & 15);
						d >>= 4;
					}
				}
			}
		}
	}
	for (x = dx; x < bpl; x++) {
		for (y = 0; y < dy; y++) {
			buf[y*bpl+x] = 0;
		}
	}
	for (x = dy*bpl; x < bpi; x++) {
		buf[x] = 0;
	}

	// t@C
	fclose(fp);

	// p[^̐ݒ(̂Q)
	*pdx  = dx;
	*pdy  = dy;
	*pbpl = bpl;
}

void WriteSprite(char *name, unsigned char *buf, short dx, short dy, short bpl) {
	FILE *fp;
	static int tblx[] = {0, 0, 8, 8};
	static int tbly[] = {0, 8, 0, 8};
	unsigned char tmp[128];
	unsigned char *p;
	unsigned long d;
	int x, y, tx, ty, w, h, i;

	// t@CJ
	fp = fopen(name, "wb");
	if (fp == NULL) {
		printf("Error: can't open %s!\n", name);
		return;
	}

	// f[^ϊď
	for (y = 0; y < dy; y += 16) {
		for (x = 0; x < dx; x += 16) {
			p = tmp;
			for (i = 0; i < 4; i++) {
				tx = tblx[i];
				ty = tbly[i];
				for (h = 0; h < 8; h++) {
					d = 0;
					for (w = 0; w < 8; w++) {
						d <<= 4;
						d  |= buf[(y+ty+h)*bpl+(x+tx+w)];
					}
					*(p  ) = (unsigned char)((d >> 24) & 0xff);
					*(p+1) = (unsigned char)((d >> 16) & 0xff);
					*(p+2) = (unsigned char)((d >>  8) & 0xff);
					*(p+3) = (unsigned char)((d      ) & 0xff);
					p += 4;
				}
			}
			fwrite(tmp, sizeof(char), 128, fp);
		}
	}

	// t@C
	fclose(fp);
}

void ReadBitmap(char *name, unsigned char *buf, short *pdx, short *pdy, short *pbpl) {
	FILE *fn;
	unsigned char src[1024];
	long offset;
	int w, h, bpl, plane, bits, comp, color;
	int x, y, i;

	// ̓t@CJ
	fn = fopen(name, "rb");
	if (fn == NULL) {
		printf("Error: can't open %s!\n", name);
		exit(1);
	}

	// wb_̓ǂݍ
	fread(src, sizeof(char), 54, fn);

	// wb_̉
	offset = (((int)src[10])      ) | (((int)src[11]) <<  8) |
             (((int)src[12]) << 16) | (((int)src[13]) << 24);
	w      = (((int)src[18])      ) | (((int)src[19]) <<  8);
	h      = (((int)src[22])      ) | (((int)src[23]) <<  8);
	plane  = (((int)src[26])      );
	bits   = (((int)src[28])      );
	comp   = (((int)src[30])      );
	color  = (((int)src[46])      );

	if ((plane != 1) && (bits != 4) && (comp != 0)) {
		printf("Error: not 4bits noncompression Windows Bitmap. %s!\n", name);
		exit(1);
	}
	if (color == 0) color = 1 << bits;

	// pbg̓ǂݍ
	if (md_palette == MD_BMP_FILE) {
		fread(src, sizeof(char), 4*color, fn);
		x = 0;
		for (i = 0; i < color; i++) {
			blue[i]  = src[x++];
			green[i] = src[x++];
			red[i]   = src[x++];
			x++;
		}
		while (i < 16) {
			blue[i]  = 0;
			green[i] = 0;
			red[i]   = 0;
			i++;
		}
	}

	// sNZf[^̓ǂݍ
	fseek(fn, offset, SEEK_SET);
	bpl = ((w + 7) & 0xfffffff8) + 2;
	for (y = h-1; y >= 0; y--) {
		fread(src, sizeof(char), ((w+7)>>1)&0xfffffffc, fn);
		i = 0;
		for (x = 0; x < w; x += 2) {
			buf[y*bpl+x  ] = (src[i] >> 4) & 15;
			buf[y*bpl+x+1] = (src[i]     ) & 15;
			i++;
		}
	}

	// t@C
	fclose(fn);

	// p[^̐ݒ
	*pdx  = (w + 7) & 0xfffffff8;
	*pdy  = h;
	*pbpl = bpl;
}

void WriteBitmap(char *name, unsigned char *buf, short dx, short dy, short bpl) {
	FILE *fn;
	unsigned char dst[1024];
	long size;
	long offset;
	int x, y, i;

	dx = (dx + 7) & 0xfffffff8;

	// o̓t@CJ
	fn = fopen(name, "wb");
	if (fn == NULL) {
		printf("Error: can't open %s!\n", name);
		exit(1);
	}

	// wb_̏
	for (i = 0; i < 128; i++) dst[i] = 0;

	offset  = 14;           // sizeof(BITMAPFILEHEADER);
	offset += 40;           // sizeof(BITMAPINFOHEADER);
	offset += 4*16;         // sizeof(RGBQUAD) * 16;
	size    = offset;
	size   += dx * dy / 2;  // width * height * 4 / 8

	// BITMAPFILEHEADER
	dst[ 0] = 'B';
	dst[ 1] = 'M';
	dst[ 2] = (size        ) & 0xff;
	dst[ 3] = (size   >>  8) & 0xff;
	dst[ 4] = (size   >> 16) & 0xff;
	dst[ 5] = (size   >> 24) & 0xff;
	dst[10] = (offset      ) & 0xff;
	dst[11] = (offset >>  8) & 0xff;
	dst[12] = (offset >> 16) & 0xff;
	dst[13] = (offset >> 24) & 0xff;

	// BITMAPINFOHEADER
	dst[14] = 40;
	dst[18] = (dx     ) & 0xff;
	dst[19] = (dx >> 8) & 0xff;
	dst[22] = (dy     ) & 0xff;
	dst[23] = (dy >> 8) & 0xff;
	dst[26] = 1;  // Plane
	dst[28] = 4;  // BitPerPixel
	dst[30] = 0;  // BI_RGB

	// pbg
	x = 54;
	for (i = 0; i < 16; i++) {
		dst[x++] = (unsigned char)blue[i];
		dst[x++] = (unsigned char)green[i];
		dst[x++] = (unsigned char)red[i];
		dst[x++] = 0;
	}

	// 
	fwrite(dst, sizeof(char), 118, fn);

	// sNZf[^̏
	for (y = dy-1; y >= 0; y--) {
		i = 0;
		for (x = 0; x < dx; x += 2) {
			dst[i++] = (unsigned char)(
				(buf[y*bpl+x  ] << 4) |
				(buf[y*bpl+x+1] & 15)
			);
		}
		fwrite(dst, sizeof(char), i, fn);
	}

	// t@C
	fclose(fn);
}

void ReadVram(short mode, unsigned char *buf, short *pdx, short *pdy, short *pbpl) {
#ifdef WIN32
	printf("Error: vram read not saport for WIN32!\n");
	exit(1);
#else
	short *src1, *src2;
	char  *dst1, *dst2;
	int x, y, w, h, sbpl, dbpl;

	switch (mode) {
	case 0:
		src1 = (short*)(void*)0xC00000;
		w    = 512;
		h    = 512;
		break;
	case 1:
		src1 = (short*)(void*)0xC00000;
		w    = 256;
		h    = 256;
		break;
	case 2:
		src1 = (short*)(void*)0xC00200;
		w    = 256;
		h    = 256;
		break;
	case 3:
		src1 = (short*)(void*)0xC40000;
		w    = 256;
		h    = 256;
		break;
	case 4:
		src1 = (short*)(void*)0xC40200;
		w    = 256;
		h    = 256;
		break;
	}
	sbpl = 512;
	dbpl = w + 2;

	// sNZf[^̓]
	dst1 = (char*)buf;
	for (y = 0; y < h; y++) {
		src2 = src1;
		dst2 = dst1;
		for (x = 0; x < w; x++) {
			*(dst2++) = *(src2++) & 15;
		}
		src1 += sbpl;
		dst1 += dbpl;
	}

	*pdx  = w;
	*pdy  = h;
	*pbpl = dbpl;
#endif
}

void WriteVram(short mode, unsigned char *buf, short dx, short dy, short bpl) {
#ifdef WIN32
	printf("Error: write vram not saport for WIN32!\n");
	exit(1);
#else
	char  *src1, *src2;
	short *dst1, *dst2;
	int x, y, w, h, sbpl, dbpl;

	switch (mode) {
	case 0:
		dst1 = (short*)(void*)0xC00000;
		w    = (dx > 512) ? 512 : dx;
		h    = (dx > 512) ? 512 : dy;
		break;
	case 1:
		dst1 = (short*)(void*)0xC00000;
		w    = (dx > 256) ? 256 : dx;
		h    = (dy > 256) ? 256 : dy;
		break;
	case 2:
		dst1 = (short*)(void*)0xC00200;
		w    = (dx > 256) ? 256 : dx;
		h    = (dy > 256) ? 256 : dy;
		break;
	case 3:
		dst1 = (short*)(void*)0xC40000;
		w    = (dx > 256) ? 256 : dx;
		h    = (dy > 256) ? 256 : dy;
		break;
	case 4:
		dst1 = (short*)(void*)0xC40200;
		w    = (dx > 256) ? 256 : dx;
		h    = (dy > 256) ? 256 : dy;
		break;
	}
	sbpl = bpl;
	dbpl = 512;

	// pbgf[^̓]
	dst2 = (short*)(void*)0xE82000;
	for (x = 0; x < 16; x++) {
		*(dst2++) = (short)(
			((red[x]   & 0xf8) << 3) |
			((green[x] & 0xf8) << 8) |
			((blue[x]  & 0xf8) >> 2)
		);
	}

	// sNZf[^̓]
	src1 = (char*)buf;
	for (y = 0; y < h; y++) {
		src2 = src1;
		dst2 = dst1;
		for (x = 0; x < w; x++) {
			*(dst2++) = *(src2++);
		}
		src1 += sbpl;
		dst1 += dbpl;
	}
#endif
}

#ifdef WIN32
struct  _finddata_t  findinfo;
long  findhandle;
#else
struct  _filbuf  findinfo;
#endif
char  finddir[256];
char  findname[256];

char *FindFirst(char *dir, char *ext) {
	char spec[256];
	char *p1, *p2;
#ifndef WIN32
	int r;
#endif

	p1 = dir;
	p2 = finddir;
	while (*p1) *(p2++) = *(p1++);
	p2--;
	if (*p2 != '\\') p2++;
	*(p2++) = '\\';
	*p2 = '\0';

	strcpy(spec, finddir);
	strcat(spec, ext);

#ifdef WIN32
	findhandle = _findfirst(spec, &findinfo);

	if (findhandle < 0) return NULL;
#else
	r = _dos_files(&findinfo, spec, 0x2b);

	if (r < 0) return NULL;
#endif
	strcpy(findname, finddir);
	strcat(findname, findinfo.name);

	return findname;
}

char *FindNext() {
	int r;

#ifdef WIN32
	r = _findnext(findhandle, &findinfo);
#else
	r = _dos_nfiles(&findinfo);
#endif
	if (r < 0) return NULL;

	strcpy(findname, finddir);
	strcat(findname, findinfo.name);

	return findname;
}

void FindClose() {
#ifdef WIN32
	_findclose(findhandle);
#endif
}

char outfilename[256];

char *NextFileName(char *infile, char *outfile) {
	char *ret, *p1, *p2, *p3;

	if (md_output == MD_SP_DIR) {
		p1 = outfile;
		p2 = outfilename;
		while (*p1) *(p2++) = *(p1++);
		p2--;
		if (*p2 != '\\') p2++;
		*(p2++) = '\\';

		p1 = infile;
		p3 = p1;
		while (*p1) {
			if (*(p1++) == '\\') p3 = p1;
		}

		while (*p3 && (*p3 != '.')) *(p2++) = *(p3++);
		*(p2++) = '.';
		*(p2++) = 's';
		*(p2++) = 'p';
		*(p2++) = '\0';

		ret = outfilename;
	} else if (md_output == MD_BMP_DIR) {
		p1 = outfile;
		p2 = outfilename;
		while (*p1) *(p2++) = *(p1++);
		p2--;
		if (*p2 != '\\') p2++;
		*(p2++) = '\\';

		p1 = infile;
		p3 = p1;
		while (*p1) {
			if (*(p1++) == '\\') p3 = p1;
		}

		while (*p3 && (*p3 != '.')) *(p2++) = *(p3++);
		*(p2++) = '.';
		*(p2++) = 'b';
		*(p2++) = 'm';
		*(p2++) = 'p';
		*(p2++) = '\0';

		ret = outfilename;
	} else {
		ret = outfile;
	}

	return ret;
}
